#!/usr/bin/python

# PageMac GPS Raspberry Pi 10 Hz GPS / IMU Data Logger
# https://www.pagemac.com/projects/rpi_gps_imu
# written by Daniel Smith, 2015-2019

# this code is Python 2.7

import io
import os
import time
import threading
import thread
import signal
import sys
import subprocess
import serial
import re
from datetime import datetime
import pytz
from Adafruit_LSM303 import Adafruit_LSM303

from adafruitOLED import adafruitOLED

# select log file location
#LOG_PATH = "/home/pi/"
LOG_PATH = "/mnt/usbdisk/"

bGPSlock = -1

class navData:
	bLogFileOpen = 0
	logFilePath = ""
	logFile = ""
	
	def __init__(self):
		timestamp = "0"
		longitude = "0.0 W"
		latitude = "0.0 N"
		
		gps_heading = "0"
		altitude = "0"			# MSL, in meters
		speed = "0"				# knots
		
		roll = "0"
		pitch = "0"
		heading = "0"
		
		extra = ""
		
	def startLogFile(self, filename):
		global LOG_PATH
		self.logFilePath = LOG_PATH + filename
		print ">Creating log " + self.logFilePath
		self.logFile = open(self.logFilePath,'a+')	# create a new file
		self.bLogFileOpen = 1
				
	def getNavData(self):
		# UTC timestamp, longitude (dd mm.mmmm E/W), latitude (ddd mm.mmmm N/S), 
		# altitude (meters), ground speed (knots), GPS heading, roll, pitch, IMU heading, 
		# accelerometer x, accelerometer y, accelerometer z, magnometer x, magnometer y, magnometer z
		data = []
		data.append(self.timestamp)
		data.append(self.longitude)
		data.append(self.latitude)
		data.append(self.altitude)
		data.append(self.speed)
		data.append(self.gps_heading)
		data.append("{0:0.2f}".format(self.roll))
		data.append("{0:0.2f}".format(self.pitch))
		data.append("{0:0.2f}".format(self.heading))
		if self.extra != "":
			data.append(self.extra)
		return ",".join(data) + "\r\n"
	
	def writeNavData(self):
		if self.bLogFileOpen == 1:
			navData = self.getNavData()
			#print self.extra
			self.logFile.write(navData)

		
class gpsSerial:
	serialNextDataDibsFlag = 0
	serialNextData = ""
	
	def __init__(self, device, baudrate, processDataFunction):
		self.device = device
		self.baudrate = baudrate
		self.processDataFunction = processDataFunction
		self.portOpen = False
		self.serialObj = serial.Serial(self.device, self.baudrate, timeout=0.05)
	
	def open(self):
		self.serialObj.close()
		self.serialObj.open()
		self.portOpen = self.serialObj.isOpen() 

	def close(self):
		self.portOpen = False
		self.serialObj.close()
		
	def readSerialThread(self):	# thread runs in the background to constantly read data from the serial port
		while (self.portOpen):	# do this forever
			line = ""
			try:
				#data = self.serialObj.read(1000);				#read serial port data
				line = self.serialObj.readline()
			except KeyboardInterrupt:				# on CTL-C
				print " - Keyboard interrupt!!!"
				break
			except Exception as e:
				print " - readSerial error: %r" % e
				pass

			if line != "":
				thread.start_new_thread(self.processDataFunction, (line,))	# process the data in another thread

	def sendLine(self, cmd):		# write a line to serial, and eppend \r\n
		self.serialObj.write(cmd + "\r\n")
		
def NMEAchecksum(sentence):
    # grab all characters between '$' and '*' and removes newline characters
	cksum = sentence[len(sentence) - 2:]
	chksumdata = re.sub("(\n|\r\n)","", sentence[sentence.find("$")+1:sentence.find("*")])
    
	# init first XOR value
	csum = 0

	for c in chksumdata:
		csum ^= ord(c)

	if csum == int(cksum, 16):
		return True
	else:
		return False

def processNMEAline(line):
	#global logFile
	global bGPSlock
	global myNavData
	global myIMU
	
	line = line.replace("\r\n", "")
	
	gps_data = line.split(",")
	if len(gps_data) < 10:
		return "0"
	if NMEAchecksum(line) == False:
		return "0"
	
	if gps_data[0] == "$GPRMC":
		if gps_data[2] == "A":
			lock = 1
			if myNavData.bLogFileOpen == 0:	# create the log file
				# format the UTC date/time data from the GPS into a filename
				# yyyy-mm-dd_hhmmss.csv
				myNavData.startLogFile("20" + gps_data[9][4:6] + "-" + gps_data[9][2:4] + "-" + gps_data[9][0:2] + "_" + gps_data[1].split(".")[0] + ".csv")
		else:
			lock = 0
		
		if bGPSlock == -1:
			if lock == 1:
				print ">GPS locked"
			else:
				print ">GPS not locked"
			bGPSlock = lock
		if lock == 1 and bGPSlock == 0:
			print ">Acquired GPS"
		elif lock == 0 and bGPSlock == 1:
			print ">Lost GPS"
		
		
		bGPSlock = lock
	
	if bGPSlock == 1:
		display.mode = 1
		if gps_data[0] == "$GPRMC" and len(gps_data) == 13:
			myNavData.speed = gps_data[7]
			
			myNavData.gps_heading = gps_data[8]	
			
			display.speed =  int(round(float(myNavData.speed)*1.15078))	# convert knots to mph
			
		if gps_data[0] == "$GPGGA" and len(gps_data) == 15:
			myNavData.timestamp = gps_data[1]
					
			myNavData.latitude = gps_data[2][:2] + " " + gps_data[2][2:] + " " + gps_data[3]
			myNavData.longitude = gps_data[4][:3] + " " + gps_data[4][3:] + " " + gps_data[5]
			myNavData.altitude = gps_data[9]
			
			IMU_data = myIMU.read()
			myNavData.roll = IMU_data[2][0]
			myNavData.pitch	= IMU_data[2][1]
			myNavData.heading = IMU_data[2][2]
			
			myNavData.extra = ",".join(str(s) for s in IMU_data[0]) + "," + ",".join(str(s) for s in IMU_data[1])		# add raw accelerometer and magnometer data
			
			myNavData.writeNavData()
	else:
		display.displayText("No GPS")
		display.mode = 0
			
			
		
print ">XGF startup"

if not os.path.isdir(LOG_PATH):
	print ">ERROR, log directory '" + LOG_PATH + "' does not exist!"
	LOG_PATH = "/home/pi/"
	#exit()

myNavData = navData()
display = adafruitOLED()

print ">Init IMU"
myIMU = Adafruit_LSM303()
myIMU.setMagGain(myIMU.LSM303_MAGGAIN_8_1)
		
print ">Init GPS"
mySerial = gpsSerial("/dev/ttyAMA0", 9600, processNMEAline)
mySerial.open()
time.sleep(.5)
print ">Configuring GPS"
mySerial.sendLine("$PMTK251,115200*1F\r\n"); # tell the gps to switch to 115200 baud 
mySerial.close()
mySerial.serialObj.baudrate = 115200	# change our serial port baud rate to 115200
mySerial.baudrate = 115200
mySerial.open()
thread.start_new_thread(mySerial.readSerialThread, ())	# start a thread to read the serial port constantly
time.sleep(.5)

mySerial.sendLine("$PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*28")	# tell the GPS to print GPRMC and GGA outputs only
#mySerial.sendLine("$PMTK220,1000*1F")	# set GPS update rate to 1 Hz
mySerial.sendLine("$PMTK220,100*2F") # set GPS update rate to 10 Hz



exitFlag = 0


print ">Main loop"
try:
	while 1:
		display.displayMode()
		if exitFlag == 1:
			break
		time.sleep(.1)

except KeyboardInterrupt:	
	print " - Keyboard interrupt!"
except Exception as e:
	print " - Main except: %r" % e
	
try:
	if myNavData.bLogFileOpen == 1:
		myNavData.logFile.close() 
	mySerial.close()
	#printMsg("Exiting... " + str(sys.exc_info()[0]))
except Exception as e:
	print " - Exit except: %r" % e
os._exit(1)
